from math import factorial
from typing import Tuple

import casadi as cs
import numpy as np
from scipy.linalg import expm
from scipy.signal import cont2discrete


def forward_euler(
    A: np.ndarray, B: np.ndarray, ts: float, c: np.ndarray = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Discretise the continuous time system x_dot = Ax + Bu using forward euler method.
    If c also passed uses system x_dot = Ax + Bu + c"""
    Ad = np.eye(A.shape[0]) + ts * A
    Bd = ts * B
    if c is None:
        return Ad, Bd
    else:
        cd = ts * c
        return Ad, Bd, cd


def zero_order_hold(
    A: np.ndarray, B: np.ndarray, ts: float, N: int = 30
) -> Tuple[np.ndarray, np.ndarray]:
    """Discretise the continuous time system x_dot = Ax + Bu using ZOH"""
    n = A.shape[0]
    I = np.eye(n)
    if isinstance(A, np.ndarray):
        D = expm(ts * np.vstack((np.hstack([A, I]), np.zeros((n, 2 * n)))))
        Ad = D[:n, :n]
        Id = D[:n, n:]
        Bd = Id.dot(B)
    else:
        M = ts * cs.vertcat(cs.horzcat(A, I), np.zeros((n, 2 * n)))
        D = sum(cs.mpower(M, k) / factorial(k) for k in range(N))
        Ad = D[:n, :n]
        Id = D[:n, n:]
        Bd = Id @ B
    return Ad, Bd


def tustin(A: np.ndarray, B: np.ndarray, ts: float) -> Tuple[np.ndarray, np.ndarray]:
    """Discretise the continuous time system x_dot = Ax + Bu using ZOH"""
    C = np.eye(A.shape[0])
    D = np.array([[0]])
    Ad, Bd, Cd, Dd, _ = cont2discrete((A, B, C, D), ts, method="bilinear")
    return Ad, Bd



import cvxpy as cp
import numpy as np

n = 2
n_l = 2

Q1 = np.array([[1, 2], [2, 4]])
Q2 = np.array([[1, 0.2], [0.2, 0.5]])
A1 = 10 * np.random.rand(1, n * n_l) - 5
A2 = 10 * np.random.rand(1, n * n_l) - 5
b1 = -5 * np.random.rand(1)
b2 = -5 * np.random.rand(1)

x1 = cp.Variable((n, 1))
x2 = cp.Variable((n, 1))

cost = cp.QuadForm(x1, Q1) + cp.QuadForm(x2, Q2)
objective = cp.Minimize(cost)

constraints = [A1 @ cp.vstack([x1, x2]) <= b1, A2 @ cp.vstack([x1, x2]) <= b2]

problem = cp.Problem(objective, constraints)

problem.solve()

print("x1 = " + str(x1.value))
print("x2 = " + str(x2.value))

print("d1 = " + str(constraints[0].dual_value))
print("d2 = " + str(constraints[1].dual_value))

# ADMM



from typing import Literal, Optional, TypeVar

import casadi as cs
import numpy as np
from csnlp.wrappers import Mpc
from csnlp.wrappers.wrapper import Nlp

SymType = TypeVar("SymType", cs.SX, cs.MX)


class MpcAdmm(Mpc[cs.SX]):
    """The local MPC minimisation in an ADMM scheme."""

    def __init__(
        self,
        nlp: Nlp[SymType],
        prediction_horizon: int,
        control_horizon: Optional[int] = None,
        input_spacing: int = 1,
        shooting: Literal["single", "multi"] = "multi",
    ) -> None:
        """Initialises the ADMM based MPC

        Parameters
        ----------
        nlp : Nlp
            NLP scheme to be wrapped
        prediction_horizon : int
            A positive integer for the prediction horizon of the MPC controller.
        control_horizon : int, optional
            A positive integer for the control horizon of the MPC controller. If not
            given, it is set equal to the control horizon.
        input_spacing : int, optional
            Spacing between independent input actions. This argument allows to reduce
            the number of free actions along the control horizon by allowing only the
            first action every `n` to be free, and the following `n-1` to be fixed equal
            to that action (where `n` is given by `input_spacing`). By default, no
            spacing is allowed, i.e., 1.
        shooting : 'single' or 'multi', optional
            Type of approach in the direct shooting for parametrizing the control
            trajectory. See [1, Section 8.5]. By default, direct shooting is used.

        Raises
        ------
        ValueError
            Raises if the shooting method is invalid; or if any of the horizons are
            invalid."""
        self._fixed_pars_init = {}

        super().__init__(
            nlp, prediction_horizon, control_horizon, input_spacing, shooting
        )
        self.N = prediction_horizon

    @property
    def fixed_pars_init(self) -> int:
        """Gets the prediction horizon of the MPC controller."""
        return self._fixed_pars_init

    @fixed_pars_init.setter
    def fixed_pars_init(self, value):
        """Prevent setting the dictionary, as it must contain the init values for y and z"""
        raise ValueError(
            "Can't set the value of fixed_pars_init. You can only add too it."
        )

    def augmented_state(self, num_neighbours, my_index, size: int = 1):
        """Generates the local state and variables for the copies of neighbour
        states over the prediction horizon

        Parameters
        ----------
        num_neighbours
            Number of neighbours coupled to this agent.
        my_index
            The index of the agents local state in the augmented state.
        size
            Dimension of local state.
        """

        x, _ = self.state("x", size)  # local state
        x_c, _, _ = self.variable(
            "x_c", (size * (num_neighbours), self.N + 1)
        )  # neighbor states

        # adding them together as the full decision var for augmented cost
        self.x_cat = cs.vertcat(
            x_c[: (my_index * size), :], x, x_c[(my_index * size) :, :]
        )

        # Parameters in augmented lagrangian
        self.y = self.parameter("y", (size * (num_neighbours + 1), self.N + 1))
        self.z = self.parameter("z", (size * (num_neighbours + 1), self.N + 1))
        self._fixed_pars_init["y"] = np.zeros((size * (num_neighbours + 1), self.N + 1))
        self._fixed_pars_init["z"] = np.zeros((size * (num_neighbours + 1), self.N + 1))

        return x, x_c

    def set_dynamics(
        self,
        F,
        n_in,
        n_out,
    ) -> None:
        raise RuntimeError(
            "For ADMM based MPC the dynamics must be set manually as constraints due to the coupling."
        )

    def set_local_cost(self, local_cost: cs.SX, rho: float = 0.5):
        """Sets the cost function for the ADMM based MPC. The augmented lagrangian
        terms are augmented to the local cost."""

        if not hasattr(self, "x_cat"):
            raise RuntimeError(
                "You must call augmented_state before setting local cost."
            )

        self.minimize(
            local_cost
            + sum(
                (self.y[:, [k]].T @ (self.x_cat[:, [k]] - self.z[:, [k]]))
                for k in range(self.N + 1)
            )
            + sum(
                ((rho / 2) * cs.norm_2(self.x_cat[:, [k]] - self.z[:, [k]]) ** 2)
                for k in range(self.N + 1)
            )
        )



import unittest

from rldmpc.cvxpy_system.evaluation import LtiNetwork


class TestStageCost(unittest.TestCase):
    def test_without_violation(self):
        LtiNetwork()

    def test_with_violation(self):
        pass


if __name__ == "__main__":
    unittest.main()



from copy import deepcopy
from typing import Any, Collection, Dict, Literal, Optional, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from csnlp.wrappers import Mpc
from gymnasium import Env
from mpcrl import LstdQLearningAgent
from mpcrl.agents.agent import ActType, ObsType, SymType
from mpcrl.agents.lstd_q_learning import ExpType
from mpcrl.agents.rl_learning_agent import LrType
from mpcrl.core.experience import ExperienceReplay
from mpcrl.core.exploration import ExplorationStrategy, StepWiseExploration
# from mpcrl.core.learning_rate import LearningRate
from mpcrl.core.parameters import LearnableParametersDict
from mpcrl.core.schedulers import Scheduler
from mpcrl.core.update import UpdateStrategy
from mpcrl.wrappers.agents import RecordUpdates

from dmpcrl.core.admm import AdmmCoordinator
from dmpcrl.core.consensus import ConsensusCoordinator
from dmpcrl.mpc.mpc_admm import MpcAdmm


class LstdQLearningAgentCoordinator(LstdQLearningAgent):
    """Coordinator to handle the communication and learning of a multi-agent q-learning system."""

    admm_iters = 50
    consensus_iters = 100

    def __init__(
        self,
        mpc_cent: Mpc[SymType],
        update_strategy: Union[int, UpdateStrategy],
        discount_factor: float,
        # learning_rate: Union[LrType, Scheduler[LrType], LearningRate[LrType]],    # TODO find out why learning rate cannot import
        learning_rate,
        learnable_parameters: LearnableParametersDict[SymType],
        n: int,
        mpc_dist_list: list[MpcAdmm],
        learnable_dist_parameters_list: list[LearnableParametersDict[SymType]],
        fixed_dist_parameters_list: list,
        G: list[list[int]],
        Adj: np.ndarray,
        rho: float,
        fixed_parameters: Union[
            None, Dict[str, npt.ArrayLike], Collection[Dict[str, npt.ArrayLike]]
        ] = None,
        exploration: Optional[ExplorationStrategy] = None,
        experience: Optional[ExperienceReplay[ExpType]] = None,
        max_percentage_update: float = float("+inf"),
        warmstart: Literal["last", "last-successful"] = "last-successful",
        hessian_type: Literal["approx", "full"] = "approx",
        record_td_errors: bool = False,
        cho_maxiter: int = 1000,
        cho_solve_kwargs: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        centralised_flag: bool = False,
        centralised_debug: bool = False,
    ) -> None:
        """Instantiates the coordinator. If centralised_flag is true it acts as a LstdQLearningAgent."""
        # TODO parameter descriptions
        self.centralised_flag = centralised_flag
        self.centralised_debug = centralised_debug
        self.n = n

        super().__init__(  # use itself as a learning agent for error checking
            mpc_cent,
            update_strategy,
            discount_factor,
            learning_rate,
            learnable_parameters,
            fixed_parameters,
            exploration,
            experience,
            max_percentage_update,
            warmstart,
            hessian_type,
            record_td_errors,
            cho_maxiter,
            cho_solve_kwargs,
            name,
        )

        if not centralised_flag:  # act as a coordinator of learning agents
            self._updates_enabled = False  # Turn updates off for the centralised
            # configure copies of exploration with different rndom generators to avoid all agents exploring identicaly
            exploration_list = [None] * n
            if exploration is not None:
                for i in range(n):
                    new_exp = deepcopy(exploration)
                    new_exp.reset(i)  # reseting with new seed
                    exploration_list[i] = StepWiseExploration(
                        new_exp, self.admm_iters, stepwise_decay=False
                    )  # step wise to account for ADMM iters
            self.agents: list[LstdQLearningAgent] = []
            for i in range(n):
                self.agents.append(  # create agents here, passing the mpc, learnable, and fixed params from the lists
                    RecordUpdates(
                        LstdQLearningAgent(
                            mpc_dist_list[i],
                            deepcopy(update_strategy),
                            discount_factor,
                            deepcopy(learning_rate),
                            learnable_dist_parameters_list[i],
                            fixed_dist_parameters_list[i],
                            exploration_list[i],
                            deepcopy(experience),
                            max_percentage_update,
                            warmstart,
                            hessian_type,
                            record_td_errors,
                            cho_maxiter,
                            cho_solve_kwargs,  # TODO add copy
                            f"{name}_{i}",
                        )
                    )
                )

            # vars for admm procedures
            self.N = mpc_cent.horizon

            self.nx_l = mpc_dist_list[
                0
            ].nx_l  # used to seperate states into each agents
            self.nu_l = mpc_dist_list[
                0
            ].nu_l  # used to seperate control into each agents

            self.admm_coordinator = AdmmCoordinator(
                self.agents,
                G,
                N=mpc_cent.horizon,
                nx_l=self.nx_l,
                nu_l=self.nu_l,
                rho=rho,
                iters=self.admm_iters,
            )

            self.consensus_coordinator = ConsensusCoordinator(Adj, self.consensus_iters)

    # need to override train method to manually reset all agent MPCs and NOT activate updates for centralised
    def train(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        seed=None,
        raises: bool = True,
        env_reset_options: Optional[dict[str, Any]] = None,
    ):
        if not self.centralised_flag:
            for agent in self.agents:
                agent.reset(seed)
            self._updates_enabled = False
        else:
            self._updates_enabled = True

        self._raises = raises
        returns = np.zeros(episodes, float)
        self.on_training_start(env)
        seeds = map(int, np.random.SeedSequence(seed).generate_state(episodes))

        for episode, current_seed in zip(range(episodes), seeds):
            self.reset(current_seed)
            state, _ = env.reset(seed=current_seed, options=env_reset_options)
            self.on_episode_start(env, episode)
            returns[episode] = self.train_one_episode(env, episode, state, raises)
            self.on_episode_end(env, episode, returns[episode])

        self.on_training_end(env, returns)
        return returns

    def evaluate(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        deterministic: bool = True,
        seed=None,
        raises: bool = True,
        env_reset_options: Optional[Dict[str, Any]] = None,
    ):
        if self.centralised_flag:
            return super().evaluate(env, episodes, seed, raises, env_reset_options)

        for agent in self.agents:
            agent.reset(seed)

        returns = np.zeros(episodes)
        self.on_validation_start(env)
        seeds = map(int, np.random.SeedSequence(seed).generate_state(episodes))

        for episode, current_seed in zip(range(episodes), seeds):
            self.reset(current_seed)
            state, _ = env.reset(seed=current_seed, options=env_reset_options)
            truncated, terminated, timestep = False, False, 0
            self.on_episode_start(env, episode)

            while not (truncated or terminated):
                joint_action, sol_list, error_flag = self.distributed_state_value(
                    state, episode, raises, deterministic=deterministic
                )
                if error_flag:
                    return returns

                state, r, truncated, terminated, _ = env.step(joint_action)
                self.on_env_step(env, episode, timestep)

                returns[episode] += r
                timestep += 1
                self.on_timestep_end(env, episode, timestep)

            self.on_episode_end(env, episode, returns[episode])

        self.on_validation_end(env, returns)
        return returns

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        truncated = terminated = False
        timestep = 0
        rewards = 0.0
        state = init_state

        if self.centralised_flag:
            return super().train_one_episode(env, episode, init_state, raises)
        else:
            # solve for the first action
            action, solV = self.state_value(
                state, False
            )  # get centralised result for comparrison
            if not solV.success:
                self.on_mpc_failure(episode, None, solV.status, raises)

            joint_action, solV_list, error_flag = self.distributed_state_value(
                state, episode, raises, deterministic=False
            )

            iter_count = 0
            while not (truncated or terminated):
                iter_count += 1
                print(iter_count)
                # compute Q(s,a)
                solQ_list, error_flag = self.distributed_action_value(
                    state, joint_action, episode, raises
                )
                if error_flag:
                    # self.on_training_end(env, rewards)
                    return rewards

                # compare dual vars of centralised and distributed sols
                # dynamics
                if self.centralised_debug:
                    solQ = self.action_value(
                        state, joint_action
                    )  # centralized val from joint action
                    cent_duals_dyn = solQ.dual_vals["lam_g_dyn"]
                    dist_duals_list_dyn: list = []
                    cent_duals_list_dyn: list = []
                    for i in range(len(self.agents)):
                        dist_duals_list_dyn.append(np.zeros((self.nx_l, self.N)))
                        cent_duals_list_dyn.append(np.zeros((self.nx_l, self.N)))
                        for k in range(self.N):  # TODO get rid of hard coded horizon
                            dist_duals_list_dyn[i][:, [k]] = np.array(
                                solQ_list[i].dual_vals[f"lam_g_dynam_{k}"]
                            )
                            cent_duals_list_dyn[i][:, [k]] = np.array(
                                cent_duals_dyn[
                                    (
                                        self.nx_l * len(self.agents) * k + self.nx_l * i
                                    ) : (
                                        self.nx_l * len(self.agents) * k
                                        + self.nx_l * (i + 1)
                                    )
                                ]
                            )
                    dynam_duals_error = np.linalg.norm(
                        cent_duals_list_dyn[0] - dist_duals_list_dyn[0]
                    )
                    if dynam_duals_error > 1e-04:
                        # exit('Duals of dynamics werent accurate!')
                        print("Duals of dynamics werent accurate!")
                # step the system with action computed at the previous iteration
                new_state, cost, truncated, terminated, info_dict = env.step(
                    joint_action
                )
                if (
                    "r_dist" in info_dict.keys()
                ):  # get distributed costs from env dict if its there
                    dist_costs = info_dict["r_dist"]
                else:
                    dist_costs = None

                self.on_env_step(env, episode, timestep)  # step centralised
                for agent in self.agents:  # step distributed agents
                    agent.on_env_step(env, episode, timestep)

                # compute V(s+) and store transition
                if self.centralised_debug:
                    new_action, solV = self.state_value(new_state, False)  # centralised
                    if not self._try_store_experience(cost, solQ, solV):
                        self.on_mpc_failure(
                            episode, timestep, f"{solQ.status}/{solV.status}", raises
                        )

                new_joint_action, solV_list, error_flag = self.distributed_state_value(
                    new_state, episode, raises, deterministic=False
                )  # distributed
                if error_flag:
                    return rewards

                # calculate centralised costs from locals with consensus
                # V_f = sum((solV_list[i].f) for i in range(len(self.agents)))
                # Q_f = sum((solQ_list[i].f) for i in range(len(self.agents)))
                V_f_vec = np.array([solV_list[i].f for i in range(len(self.agents))])
                Q_f_vec = np.array([solQ_list[i].f for i in range(len(self.agents))])
                V_f = self.consensus_coordinator.average_consensus(V_f_vec)[0] * len(
                    self.agents
                )
                Q_f = self.consensus_coordinator.average_consensus(Q_f_vec)[0] * len(
                    self.agents
                )
                if dist_costs is not None:
                    av_cost = self.consensus_coordinator.average_consensus(
                        np.asarray(dist_costs)
                    )
                    cost_f = av_cost[0] * len(self.agents)
                else:
                    cost_f = cost
                for i in range(len(self.agents)):  # store experience for agents
                    object.__setattr__(
                        solV_list[i], "f", V_f
                    )  # overwrite the local costs with the global ones TODO make this nicer
                    object.__setattr__(solQ_list[i], "f", Q_f)

                    if not self.agents[i].unwrapped._try_store_experience(
                        cost_f, solQ_list[i], solV_list[i]
                    ):
                        self.agents[i].on_mpc_failure(
                            episode,
                            timestep,
                            f"{solQ_list[i].status}/{solV_list[i].status}",
                            raises,
                        )
                # increase counters
                state = new_state
                joint_action = new_joint_action
                rewards += float(cost)
                timestep += 1

                self.on_timestep_end(env, episode, timestep)
                for agent in self.agents:
                    agent.on_timestep_end(env, episode, timestep)
            return rewards

    def distributed_state_value(self, state, episode, raises, deterministic):
        # local_action_list, local_sol_list, error_flag = self.admm(
        #    state, episode, raises, deterministic=deterministic
        # )
        (
            local_action_list,
            local_sol_list,
            error_flag,
        ) = self.admm_coordinator.solve_admm(state, deterministic=deterministic)
        if not error_flag:
            return cs.DM(local_action_list), local_sol_list, error_flag
        else:
            return None, local_sol_list, error_flag

    def distributed_action_value(self, state, action, episode, raises):
        (
            local_action_list,
            local_sol_list,
            error_flag,
        ) = self.admm_coordinator.solve_admm(state, action=action)
        return local_sol_list, error_flag



from typing import List

import casadi as cs
import numpy as np
from mpcrl import Agent


def g_map(Adj: np.ndarray):
    """Construct the mapping from local to global variables from an adjacency matrix."""
    n = Adj.shape[0]  # number of agents
    G: list[list[int]] = []
    for i in range(n):
        G.append([])
        for j in range(n):
            if Adj[i, j] == 1 or i == j:
                G[i].append(j)
    return G


class AdmmCoordinator:
    """Class for coordinating the ADMM procedure of a network of agents"""

    def __init__(
        self,
        agents: List[Agent],
        G: List[List[int]],
        N: int,
        nx_l: int,
        nu_l: int,
        rho: float,
        iters: int = 50,
    ) -> None:
        """Initiale the ADMM coordinator

        Parameters
        ----------
        agents: List[Agent]
            A list of the agents involved in the ADMM procedure.
        G: List[List[int]]
            Mapping between local and global indexes: G[i][j] = g
        N: int
            Length of prediction horizon.
        nx_l: int
            Local state dimension of agents.
        nu_l: int
            Local control dimension of agnets.
        rho: float
            Constant penalty term for augmented lagrangian.
        iters: int = 50
            Fixed number of ADMM iterations."""
        self.agents = agents
        self.n = len(agents)
        self.iters = iters
        self.G = G
        self.nx_l = nx_l
        self.nu_l = nu_l
        self.rho = rho

        # TODO confirm that agents have ADMM based MPCs

        # admm related variables
        self.y_list: list[np.ndarray] = []  # dual vars
        self.x_temp_list: list[np.ndarray] = []  # intermediate numerical values for x
        self.z = np.zeros((self.n * nx_l, N + 1))

        for i in range(self.n):
            x_dim = nx_l * len(G[i])  # dimension of augmented state for agent i
            self.y_list.append(np.zeros((x_dim, N + 1)))
            self.x_temp_list.append(np.zeros((x_dim, N + 1)))

        # generate slices of z for each agent, slices that will pull the relevant
        # component of global variable out for the agents
        z_slices: list[list[int]] = []
        for i in range(self.n):
            z_slices.append([])
            for j in self.G[i]:
                z_slices[i] += list(np.arange(nx_l * j, nx_l * (j + 1)))
        self.z_slices = z_slices

    def solve_admm(self, state, action=None, deterministic=False):
        """Solve the mpc problem for the network of agents using ADMM. If an
        action provided, the first action is constrained. Deterministic is used for
        adding exploration into the local problems."""

        loc_action_list = [None] * len(self.agents)
        local_sol_list = [None] * len(self.agents)

        for iter in range(self.iters):
            # x update: solve local minimisations

            for i in range(len(self.agents)):
                loc_state = state[
                    self.nx_l * i : self.nx_l * (i + 1), :
                ]  # get local state from global

                # set parameters in augmented lagrangian
                self.agents[i].fixed_parameters["y"] = self.y_list[i]
                self.agents[i].fixed_parameters["z"] = self.z[
                    self.z_slices[i], :
                ]  # use z slice to extract relevant z component

                if action is None:
                    loc_action_list[i], local_sol_list[i] = self.agents[i].state_value(
                        loc_state, deterministic
                    )
                else:
                    # get local action from global
                    loc_action = action[self.nu_l * i : self.nu_l * (i + 1), :]
                    local_sol_list[i] = self.agents[i].action_value(
                        loc_state, loc_action
                    )
                if not local_sol_list[i].success:  # if a local problem failed
                    return (
                        loc_action_list,
                        local_sol_list,
                        True,
                    )  # return with error flag as true

                # construct solution to augmented state from local state and coupled states
                idx = self.G[i].index(i) * self.nx_l
                self.x_temp_list[i] = cs.vertcat(
                    local_sol_list[i].vals["x_c"][:idx, :],
                    local_sol_list[i].vals["x"],
                    local_sol_list[i].vals["x_c"][idx:, :],
                )

            # z update: an averaging of all agents' optinions on each z

            for i in range(len(self.agents)):  # loop through each z
                count = 0
                sum = np.zeros((self.nx_l, self.z.shape[1]))
                for j in range(
                    len(self.agents)
                ):  # loop through agents who have opinion on this z
                    if i in self.G[j]:
                        count += 1
                        x_slice = slice(
                            self.nx_l * self.G[j].index(i),
                            self.nx_l * (self.G[j].index(i) + 1),
                        )
                        sum += self.x_temp_list[j][x_slice, :]
                self.z[self.nx_l * i : self.nx_l * (i + 1), :] = sum / count

            # y update: increment by the residual

            for i in range(len(self.agents)):
                self.y_list[i] = self.y_list[i] + self.rho * (
                    self.x_temp_list[i] - self.z[self.z_slices[i], :]
                )

        return (
            loc_action_list,
            local_sol_list,
            False,
        )  # return last solutions with error flag as false



import numpy as np


class ConsensusCoordinator:
    """Class for coordinating consensus algorithms"""

    def __init__(self, Adj: np.ndarray, iters: int = 100) -> None:
        """Instantiate coordinator.

        Parameters
        ----------
        Adj: np.ndarray
            Adjacency matrix of network.
        iters:int = 100
            Number of iterations to run conensus for."""
        self.iters = iters
        self.P = self.gen_consensus_matrix(Adj)

    def average_consensus(self, x):
        for iter in range(self.iters):
            x = self.P @ x
        return x

    def gen_consensus_matrix(self, Adj: np.ndarray):
        """Generate P matrix for consensus using graph laplacian approach."""
        max_N = max(sum(Adj))  # maximum neighbourhood cardinality
        eps = 0.5 * (1 / max_N)  # must be less than 1/max_N
        D_in = np.zeros(Adj.shape)
        for i in range(Adj.shape[0]):
            D_in[i, i] = sum(Adj)[i]
        L = D_in - Adj  # graph laplacian
        P = np.eye(Adj.shape[0]) - eps * L  # consensus matrix
        return P



from matplotlib.figure import Figure
import tikzplotlib
import matplotlib as mpl

# monkey patching to fix some issues with tikzplotlib
mpl.lines.Line2D._us_dashSeq = property(lambda self: self._dash_pattern[1])
mpl.lines.Line2D._us_dashOffset = property(lambda self: self._dash_pattern[0])
mpl.legend.Legend._ncol = property(lambda self: self._ncols)


def save2tikz(*figs: Figure) -> None:
    """Saves the figure to a tikz file (`.tex` extension). 
    See https://pypi.org/project/tikzplotlib/ for more details.

    Parameters
    ----------
    figs : matplotlib Figures
        One or more matplotlib figures to be converted to tikz files. These
        files will be named based on the number of the corresponding figure.
    """
    for fig in figs:
        tikzplotlib.save(
            f"figure_{fig.number}.tex",
            figure=fig,
            extra_axis_parameters={r"tick scale binop=\times"},
        )